import os
import os.path as osp
import numpy as np
from copy import deepcopy
from torch.utils.data import Dataset, DataLoader
from torch.utils.data.dataloader import default_collate

import utils.io as io
from utils.constants import Constants
from global_constants import hcstvg_paths


class HCSTVGDatasetConstants(Constants):
    def __init__(self,subset, det_dir):
        super().__init__()
        self.subset = subset
        self.det_dir = os.path.join(hcstvg_paths['det_dir'], det_dir,self.subset)
        self.global_feat_dir = hcstvg_paths['feat_dir']
        
        self.video_ids_txt = os.path.join(
            hcstvg_paths['proc_dir'],
            hcstvg_paths['subsets'][self.subset])
        self.sentences_json = os.path.join(
            hcstvg_paths['proc_dir'],
            hcstvg_paths['sentences'][self.subset])

        self.features_pkl = self.det_dir 
        # temporal 
        self.temporal_data = os.path.join(
            hcstvg_paths['proc_dir'],
            hcstvg_paths['temporal_data'][subset]
        )
        
        # Noun and adj tokens in captions for MI training
        self.noun_adj_tokens_json = os.path.join(
            hcstvg_paths['proc_dir'],
            hcstvg_paths['noun_adj_tokens'][subset])
        self.read_noun_adj_tokens = True
        self.max_noun_adj_tokens = 3


class HCSTVGDataset(Dataset):
    def __init__(self,const):
        super().__init__()
        self.const = deepcopy(const)
        self.video_ids = self.read_video_ids()
        # self.phrase_boxes = io.load_json_object(self.const.phrase_boxes_json)
        self.sentences = io.load_json_object(self.const.sentences_json)
        
        os.environ['HDF5_USE_FILE_LOCKING']="FALSE"
        
        if self.const.read_noun_adj_tokens is True:
            self.noun_adj_token_ids = io.load_json_object(
                self.const.noun_adj_tokens_json)
        
        self.gt_pid = io.load_json_object("test_gt_pids.json")
        # Temporal
        self.temporal_data = io.load_json_object(self.const.temporal_data)    
        self.max_num_frames_t = 256
        self.max_num_words = 20

    def read_video_ids(self):
        video_ids = io.read(self.const.video_ids_txt)
        video_ids = [idx.decode() for idx in video_ids.split()]
        return video_ids

    def __len__(self):
        return len(self.video_ids)

    def get_caption(self,video_id,cap_num):
        return self.sentences[video_id][cap_num]['sentence']

    def get_phrases(self,video_id,cap_num):
        return self.sentences[video_id][cap_num]['phrases']

    def read_object_features_hdf5(self,video_id):
        f = io.load_h5py_object(self.const.features_hdf5)
        features = f[video_id][()]
        f.close()
        return features

    def read_object_features_pkl(self,video_id):
        f = io.load_pickle_object(osp.join(self.const.features_pkl, video_id+'.pkl'))
        features = f['features']
        return features
    
    def pad_noun_adj_token_ids(self,noun_adj_token_ids):
        num_tokens = len(noun_adj_token_ids)
        if num_tokens >= self.const.max_noun_adj_tokens:
            noun_adj_token_ids = \
                noun_adj_token_ids[:self.const.max_noun_adj_tokens]
        else:
            padding = [-1]*(self.const.max_noun_adj_tokens - num_tokens)
            noun_adj_token_ids = noun_adj_token_ids + padding

        return noun_adj_token_ids
    
    
    def _sample_frame_features(self, pid_features):
        pid_feat = []
        num_clips = 32
        for pid, pid_meta in pid_features.items():
            frames_feat = np.array(pid_meta)
            keep_idx = np.arange(0, num_clips + 1) / num_clips * len(frames_feat)
            keep_idx = np.round(keep_idx).astype(np.int64)
            keep_idx[keep_idx >= len(frames_feat)] = len(frames_feat) - 1
            frames_feat1 = [frames_feat[keep_idx[j]] for j in range(num_clips)]
            pid_feat.append(np.stack(frames_feat1, 0))
        return np.array(pid_feat)
    
    def _sample_frame_features_avg(self, pid_features):
        pid_feat = []
        num_clips = 32
        for pid, pid_meta in pid_features.items():
            frames_feat = np.array(pid_meta)
            keep_idx = np.arange(0, num_clips + 1) / num_clips * len(frames_feat)
            keep_idx = np.round(keep_idx).astype(np.int64)
            keep_idx[keep_idx >= len(frames_feat)] = len(frames_feat) - 1
            frames_feat1 = []
            for j in range(num_clips):
                s, e = keep_idx[j], keep_idx[j + 1]
                assert s <= e
                if s == e:
                    frames_feat1.append(frames_feat[s])
                else:
                    frames_feat1.append(frames_feat[s:e].mean(axis=0))
            pid_feat.append(np.stack(frames_feat1, 0))
        return np.array(pid_feat)
    
    def _load_frame_features(self, vid):
        f = io.load_npy_object(osp.join(self.const.global_feat_dir, vid+'.npy'))
        return f.squeeze()
    
    @property
    def num_clips(self):
        return self.max_num_frames

    @property
    def vocab_size(self):
        return len(self.keep_vocab) + 1
    
    def _sample_frame_features_t(self, frames_feat):
        num_clips = self.max_num_frames_t
        keep_idx = np.arange(0, num_clips + 1) / num_clips * len(frames_feat)
        keep_idx = np.round(keep_idx).astype(np.int64)
        keep_idx[keep_idx >= len(frames_feat)] = len(frames_feat) - 1
        frames_feat1 = []
        for j in range(num_clips):
            s, e = keep_idx[j], keep_idx[j + 1]
            assert s <= e
            if s == e:
                frames_feat1.append(frames_feat[s])
            else:
                frames_feat1.append(frames_feat[s:e].mean(axis=0))
        return np.stack(frames_feat1, 0)
    
    def __getitem__(self,i):
        video_num = i
        cap_num = 0
        
        video_id = self.video_ids[video_num]
        caption = self.get_caption(video_id, cap_num)
        # caption = "A person playing"
        features = self.read_object_features_pkl(video_id)
        
        # Sample frames from the tube - n_objects x n_frames (32) x feat_dim
        features = self._sample_frame_features(features)  # avg

        # Ground truth for performance calculation
        gt_pid = self.gt_pid[video_id]
        
        # Temporal
        temp_data = self.temporal_data[video_id]
        duration, vid_st, vid_end, sentence  = temp_data['duration'], temp_data['start'], temp_data['end'], temp_data['sentence']
        duration = float(duration)

        frames_feat = self._sample_frame_features_t(self._load_frame_features(video_id))
        
        to_return = {
            'video_id': video_id,
            'cap_num': cap_num,
            'caption': caption,
            'tube_feat': features,
            'gt_pid': gt_pid,
            'frames_feat': frames_feat,
            'raw_caption': sentence,
            'raw': [video_id, duration, [vid_st, vid_end], sentence]
        }
        cap_id = video_id + '_' + str(cap_num)
        if self.const.read_noun_adj_tokens is True:
            noun_adj_token_ids = self.noun_adj_token_ids[cap_id]['token_ids']
            noun_adj_tokens = self.noun_adj_token_ids[cap_id]['tokens']
            to_return['noun_adj_token_ids'] = np.array(
                self.pad_noun_adj_token_ids(noun_adj_token_ids),
                dtype=np.int32)
            to_return['noun_adj_tokens'] = ' '.join(noun_adj_tokens)
        return to_return
