import os
import os.path as osp
import numpy as np
import random
from copy import deepcopy
from torch.utils.data import Dataset, DataLoader
from torch.utils.data.dataloader import default_collate

import utils.io as io
from utils.constants import Constants
from global_constants import hcstvg_paths, hcstvg_quant_paths

import pdb


class HCSTVGDatasetConstants(Constants):
    def __init__(self,subset, det_dir):
        '''
        Include detection directory change
        '''
        super().__init__()
        self.subset = subset
        self.det_dir = os.path.join(hcstvg_paths['det_dir'], det_dir,self.subset)
        self.global_feat_dir = hcstvg_paths['feat_dir']

        self.image_ids_txt = os.path.join(
            hcstvg_paths['proc_dir'],
            hcstvg_paths['subsets'][self.subset])
        self.sentences_json = os.path.join(
            hcstvg_paths['proc_dir'],
            hcstvg_paths['sentences'][self.subset])
        
        self.features_pkl = self.det_dir 
        self.gt_pid_json = os.path.join(os.path.join(hcstvg_paths['det_dir'], det_dir), f'hcstvg_{self.subset}_data_gt_class_pid.json')
        self.max_objects = 10
        self.mask_prob = 0.2

        # Noun tokens in captions for MI training
        self.noun_tokens_json = os.path.join(
            hcstvg_paths['proc_dir'],
            hcstvg_paths['noun_tokens'][self.subset])
        self.read_noun_tokens = False
        self.max_noun_tokens = 1
        
        # Noun, adj, verb tokens in captions for MI training
        self.noun_adj_tokens_json = os.path.join(
            hcstvg_paths['proc_dir'],
            hcstvg_paths['noun_adj_tokens'][subset])
        self.read_noun_adj_tokens = False
        self.max_noun_adj_tokens = 3

        self.max_num_frames = 32
        
        # temporal 
        self.temporal_data = os.path.join(
            hcstvg_paths['proc_dir'],
            hcstvg_paths['temporal_data'][subset]
        )
        
        
class HCSTVGDataset(Dataset):
    def __init__(self,const):
        super().__init__()
        self.const = deepcopy(const)
        self.image_ids = self.read_image_ids()
        
        self.sentences = io.load_json_object(self.const.sentences_json)
        
        if self.const.read_noun_tokens is True:
            self.noun_tokens = io.load_json_object(
                self.const.noun_tokens_json)
        
        if self.const.read_noun_adj_tokens is True:
            self.noun_adj_token_ids = io.load_json_object(
                self.const.noun_adj_tokens_json)
            
        self.temporal_data = io.load_json_object(self.const.temporal_data)
        self.max_num_frames_t = 256
        self.max_num_words = 20

        
    def read_image_ids(self):
        image_ids = io.read(self.const.image_ids_txt)
        image_ids = [idx.decode() for idx in image_ids.split()]
        return image_ids
            
    def __len__(self):
        return 5*len(self.image_ids)
        # return len(self.image_ids)

    def get_caption(self,image_id,cap_num):
        return self.sentences[image_id][cap_num]['sentence']

    def read_object_features_hdf5(self,image_id):
        f = io.load_h5py_object(self.const.features_hdf5)
        features = f[image_id][()]
        f.close()
        return features
    
    def read_object_features_pkl(self,image_id):
        f = io.load_pickle_object(osp.join(self.const.features_pkl, image_id+'.pkl'))
        features = f['features']
        return features
    
    def read_gt_pid(self,image_id):
        f = io.load_json_object(osp.join(self.const.gt_pid_json))
        pid = f[image_id]
        return pid
    
    def pad_tube_object_features(self,features):
        T,F,D = features.shape # num_tubes/objects x n_frames x feat. dim
        pad_mask = np.zeros(self.const.max_objects).astype(np.bool)
        if T==self.const.max_objects:
            return features, pad_mask

        if T > self.const.max_objects:
            features = features[:self.const.max_objects]
        else:
            features = np.concatenate((
                features,
                np.zeros([self.const.max_objects-T,F,D])),0).astype(np.float32)
            pad_mask[T:] = True
        return features, pad_mask
    
    def mask_objects_probablistic(self,num_objects):
        mask = np.random.uniform(size=(self.const.max_objects))
        mask[num_objects:] = 1.0
        mask = mask < self.const.mask_prob
        return mask
    
    def mask_objects(self,num_objects):
        '''
        Mask values apart from where objects are present
        '''
        mask = np.random.uniform(size=(self.const.max_objects))
        mask[:num_objects] = 1.0
        mask = mask == 1.0
        return mask
    
    def pad_noun_token_ids(self,noun_token_ids):
        num_tokens = len(noun_token_ids)
        if num_tokens >= self.const.max_noun_tokens:
            noun_token_ids = noun_token_ids[:self.const.max_noun_tokens]
        else:
            padding = [-1]*(self.const.max_noun_tokens - num_tokens)
            noun_token_ids = noun_token_ids + padding

        return noun_token_ids
    
    def pad_noun_adj_token_ids(self,noun_adj_token_ids):
        num_tokens = len(noun_adj_token_ids)
        if num_tokens >= self.const.max_noun_adj_tokens:
            noun_adj_token_ids = \
                noun_adj_token_ids[:self.const.max_noun_adj_tokens]
        else:
            padding = [-1]*(self.const.max_noun_adj_tokens - num_tokens)
            noun_adj_token_ids = noun_adj_token_ids + padding

        return noun_adj_token_ids

    def pad_noun_verb_token_ids(self,noun_verb_token_ids):
        num_tokens = len(noun_verb_token_ids)
        if num_tokens >= self.const.max_noun_verb_tokens:
            noun_verb_token_ids = \
                noun_verb_token_ids[:self.const.max_noun_verb_tokens]
        else:
            padding = [-1]*(self.const.max_noun_verb_tokens - num_tokens)
            noun_verb_token_ids = noun_verb_token_ids + padding

        return noun_verb_token_ids
    
    def get_neg_noun_samples_feats(self,image_id,cap_id,noun_id=None):
        str_image_id = str(image_id)
        str_cap_id = str(cap_id)
        
        if (str_image_id in self.neg_noun_samples) and \
            (str_cap_id in self.neg_noun_samples[str_image_id]):
            negs = self.neg_noun_samples[str_image_id][str_cap_id]['negs']
        else:
            feats = np.zeros(
                [1+self.const.num_neg_nouns,self.const.neg_noun_feat_dim],
                dtype=np.float32)
            noun_id = -1
            return feats, noun_id
        
        if noun_id is None:
            str_noun_id = random.choice(list(negs.keys()))
            noun_id = int(str_noun_id)
        
        neg_samples_feats = io.load_h5py_object(self.const.neg_noun_samples_h5py)
        feat_name = f'{str_image_id}_{str_cap_id}_{str_noun_id}'
        feats = neg_samples_feats[feat_name][()].astype(np.float32)
        neg_samples_feats.close()

        return feats, noun_id

    def _load_frame_features(self, vid):
        f = io.load_npy_object(osp.join(self.const.global_feat_dir, vid+'.npy'))
        return f.squeeze()
    
    @property
    def num_clips(self):
        return self.max_num_frames

    @property
    def vocab_size(self):
        return len(self.keep_vocab) + 1
    
    def _sample_frame_features(self, pid_features):
        pid_feat = []
        num_clips = 32
        for pid, pid_meta in pid_features.items():
            frames_feat = np.array(pid_meta)
            keep_idx = np.arange(0, num_clips + 1) / num_clips * len(frames_feat)
            keep_idx = np.round(keep_idx).astype(np.int64)
            keep_idx[keep_idx >= len(frames_feat)] = len(frames_feat) - 1
            frames_feat1 = [frames_feat[keep_idx[j]] for j in range(num_clips)]
            pid_feat.append(np.stack(frames_feat1, 0))
        return np.array(pid_feat)


    def _sample_frame_features_avg(self, pid_features):
        pid_feat = []
        num_clips = 32
        for pid, pid_meta in pid_features.items():
            frames_feat = np.array(pid_meta)
            # print(pid, frames_feat.shape)
            keep_idx = np.arange(0, num_clips + 1) / num_clips * len(frames_feat)
            keep_idx = np.round(keep_idx).astype(np.int64)
            keep_idx[keep_idx >= len(frames_feat)] = len(frames_feat) - 1
            frames_feat1 = []
            for j in range(num_clips):
                s, e = keep_idx[j], keep_idx[j + 1]
                assert s <= e
                if s == e:
                    frames_feat1.append(frames_feat[s])
                else:
                    frames_feat1.append(frames_feat[s:e].mean(axis=0))
            pid_feat.append(np.stack(frames_feat1, 0))
        # print(np.array(pid_feat).shape)
        # exit()
        return np.array(pid_feat)
    
    def _sample_frame_features_t(self, frames_feat):
        num_clips = self.max_num_frames_t
        keep_idx = np.arange(0, num_clips + 1) / num_clips * len(frames_feat)
        keep_idx = np.round(keep_idx).astype(np.int64)
        keep_idx[keep_idx >= len(frames_feat)] = len(frames_feat) - 1
        frames_feat1 = []
        for j in range(num_clips):
            s, e = keep_idx[j], keep_idx[j + 1]
            assert s <= e
            if s == e:
                frames_feat1.append(frames_feat[s])
            else:
                frames_feat1.append(frames_feat[s:e].mean(axis=0))
        return np.stack(frames_feat1, 0)

    def __getitem__(self,i):
        image_num = i//5
        cap_num = i%5
        
        # image_num = i
        # cap_num = 0
        
        image_id = self.image_ids[image_num]
        cap_id = f'{image_id}_{cap_num}'
        caption = self.get_caption(image_id, cap_num)
        features = self.read_object_features_pkl(image_id)
        
        num_objects = len(features)
        features = self._sample_frame_features(features)  # avg
        tube_features, tube_pad_mask = self.pad_tube_object_features(features)
        object_mask = self.mask_objects(num_objects)
        
        # Temporal
        temp_data = self.temporal_data[image_id]
        duration, tube_st, tube_end, sentence  = temp_data['duration'], temp_data['start'], temp_data['end'], temp_data['sentence']
        duration = float(duration)
        frames_feat = self._sample_frame_features_t(self._load_frame_features(image_id))
        
        to_return = {
            'video_id': image_id,
            'cap_num': cap_num,
            'cap_id': cap_id,
            'caption': caption,
            'tube_feat': tube_features,
            'object_mask': object_mask,
            'tube_pad_mask': tube_pad_mask,
            'num_objects': min(num_objects, 10),
            'frames_feat': frames_feat,
            'raw_caption': sentence,
            'raw': [image_id, duration, [tube_st, tube_end], sentence]
        }
        cap_id = image_id + '_' + str(cap_num)
        if self.const.read_noun_adj_tokens is True:
            noun_adj_token_ids = self.noun_adj_token_ids[cap_id]['token_ids']
            noun_adj_tokens = self.noun_adj_token_ids[cap_id]['tokens']
            to_return['noun_adj_token_ids'] = np.array(
                self.pad_noun_adj_token_ids(noun_adj_token_ids),
                dtype=np.int32)
            to_return['noun_adj_tokens'] = ' '.join(noun_adj_tokens)
            
        
        return to_return
