import os
import json
import torch
import numpy as np
import os.path as osp
import utils.io as io
from tqdm import tqdm

from utils.constants import Constants
from .models.object_encoder import ObjectEncoder
from .models.cap_encoder import CapEncoder
from exp.eval_hcstvg.dataset_videos import HCSTVGDataset

from utils.bbox_utils import compute_iou, point_in_box, compute_center
from .train import create_cap_info_nce_criterion
from .eval_hcstvg_orig_repo_new_anns_test import hcstvg_eval_new_anns_v2_temporal, hcstvg_eval_tiou

# temporal localization imports
from .models.temp_reconstruction import temp_reconstruction

def combine_tokens(tokens,words):
    num_tokens = len(tokens)
    combined_tokens = []
    i = 1
    j = 0
    # print(tokens, words)
    while(i < num_tokens-1):
        token = tokens[i]
        word = words[j].lower()
        if token==word:
            combined_tokens.append({i})
            j+=1
            i+=1
        else:
            idx_set = {i}
            prev_token=tokens[i]
            i+=1
            while(prev_token!=word):
                idx_set.add(i)
                token = tokens[i]
                if len(token) > 2 and token[:2]=='##':
                    token = token[2:]

                prev_token = prev_token + token
                i+=1
            
            combined_tokens.append(idx_set)
            j+=1
    return combined_tokens
        
def map_phrase_to_tokens(phrase_info,combined_tokens):
    # phrase = phrase_info['phrase']
    phrase = phrase_info
    # phrase = phrase_info
    num_words = min(len(phrase.split()), len(combined_tokens))
    # start_idx = phrase_info['first_word_index']
    start_idx = 0
    token_ids = set()
    for i in range(num_words):
        token_ids.update(combined_tokens[start_idx+i])
        
    return token_ids

def combine_att(token_obj_att,phrase_token_ids):
    phrase_att = None
    for i in phrase_token_ids:
        if phrase_att is None:
            phrase_att = token_obj_att[i]
        else:
            phrase_att = torch.max(phrase_att,token_obj_att[i])
        # print(phrase_att)
    return phrase_att

def select_boxes(boxes,phrase_att,k=3):
    box_ids = torch.topk(phrase_att,k)[1]
    selected_boxes = [list(map(int, boxes[i])) for i in box_ids if i < len(boxes)]
    return selected_boxes

def compute_recall(pred_boxes,gt_boxes,k=1):
    recalled = [0]*k
    pred_box = [None]*k
    gt_box = [None]*k
    for i,pred_box_ in enumerate(pred_boxes):
        if i>=k:
            break
        for gt_box_ in gt_boxes:
            iou = compute_iou(pred_box_,gt_box_)
            if iou >= 0.5:
                recalled[i] = 1
                pred_box[i] = pred_box_
                gt_box[i] = gt_box_
                break
    
    max_recall = 0
    for i in range(k):
        max_recall = max(recalled[i],max_recall)
        recalled[i] = max_recall

    return recalled, pred_box, gt_box

def compute_pt_acc(pred_boxes,gt_boxes):
    pt_recalled = False
    for i,pred_box_ in enumerate(pred_boxes):
        if i>=1:
            break
        
        pred_center = compute_center(pred_box_)
        for gt_box_ in gt_boxes:
            pt_recalled = point_in_box(pred_center,gt_box_)
            if pt_recalled is True:
                break
            
    return float(pt_recalled)

def calculate_IoU_batch2(i0, i1):
    union = (np.min(np.stack([i0[0], i1[0]], 0), 0), np.max(np.stack([i0[1], i1[1]], 0), 0))
    inter = (np.max(np.stack([i0[0], i1[0]], 0), 0), np.min(np.stack([i0[1], i1[1]], 0), 0))
    # iou = 1.0 * (inter[1] - inter[0] + 1) / (union[1] - union[0] + 1)
    iou = 1.0 * (inter[1] - inter[0] + 1e-10) / (union[1] - union[0] + 1e-10)
    iou[union[1] - union[0] < -1e-5] = 0
    iou[iou < 0] = 0.0
    return iou

def top_1_metric(pred, label):
    result = {}
    bsz = pred.shape[0]
    iou = calculate_IoU_batch2((pred[:, 0], pred[:, 1]), (label[:, 0], label[:, 1]))
    result['mIoU'] = np.mean(iou)
    for i in range(1, 10, 2):
        result['IoU@0.{}'.format(i)] = 1.0 * np.sum(iou >= i / 10) / bsz
    return result

def eval_model_lang_only_temporal(model1, model2, dataset,exp_const, data_const_subset):
    # Set mode
    model1.cap_encoder.eval()
    model1.cap2_encoder.eval()
    model1.lang_sup_criterion.eval()

    model2.temp_grounding.eval()
    
    img2boxes = {}
    print(len(dataset))
    predicted_video_durations = {}
    success_count = 0
    for it,data in enumerate(tqdm(dataset)):
        # Forward pass
        object_features = torch.FloatTensor(data['tube_feat']).cuda().unsqueeze(0)
        tube_context_object_features = object_features
        
        # Compute loss 
        # # Main caption
        token_ids, tokens, token_lens = model1.cap_encoder.tokenize_batch(
            [data['caption']], max_len=25)
        token_ids = torch.LongTensor(token_ids).cuda()
        token_features, word_word_att = model1.cap_encoder(token_ids)
        token_mask = torch.zeros(token_ids.size()).cuda()
            
        if exp_const.nav:
            noun_adj_token_ids = data['noun_adj_token_ids']
            noun_adj_token_ids = np.sort(noun_adj_token_ids[noun_adj_token_ids>=0])
            nav_token_features, nav_token_mask = model1.cap_encoder.select_embed(
                token_features,
                torch.Tensor(noun_adj_token_ids).long().cuda().unsqueeze(0))
            # pdb.set_trace()
            _, token_obj_att, _ = \
                model1.lang_sup_criterion(
                    tube_context_object_features,
                    object_features,
                    nav_token_features,
                    nav_token_mask,
                    temp_attn=exp_const.temp_attn)
            
            _, token_obj_att_caption, _ = \
                model1.lang_sup_criterion(
                    tube_context_object_features,
                    object_features,
                    token_features,
                    token_mask,
                    temp_attn=exp_const.temp_attn)
        else:
            _, token_obj_att, _ = \
                model1.lang_sup_criterion(
                    tube_context_object_features,
                    object_features,
                    token_features.float(),
                    token_mask,
                    temp_attn=exp_const.temp_attn)
            
        words = data['caption'].split()
        tokens = tokens[0]
        if exp_const.nav:
            combined_tokens = combine_tokens_na(tokens,words)
            # data['noun_adj_tokens']
            eval_token_ids = map_phrase_to_tokens(
                    ' '.join(words),
                    combined_tokens)
            eval_token_ids = sorted(list(eval_token_ids))
            eval_token_ids = torch.arange(0, len(eval_token_ids)).cuda()
            
            if len(eval_token_ids) == 0:
                # print(data['image_id'], data['caption'], data['noun_adj_tokens'], noun_adj_token_ids, 'using main caption')
                combined_tokens = combine_tokens(tokens,data['caption'].split())
                eval_token_ids = map_phrase_to_tokens(
                        data['caption'],
                        combined_tokens)
                eval_token_ids = sorted(list(eval_token_ids))
                eval_token_ids = torch.arange(0, len(eval_token_ids)).cuda()
            # pdb.set_trace()
            
            cap_att = combine_att(token_obj_att[0, 0], eval_token_ids)
            img2boxes[data['image_id']] = torch.argmax(cap_att).item()
            gt_pid = data['gt_pid']
            if gt_pid == img2boxes[data['image_id']]:
                success_count += 1
            
        videos = data['raw'][0]
        durations = np.asarray(data['raw'][1])
        
        frames_feat = torch.from_numpy(data['frames_feat']).cuda()
        words_feat, words_len, weights, words_id = model1.cap_encoder.tg_tokenize_batch([data['raw_caption']])
        output = model2.temp_grounding(frames_feat.unsqueeze(0), words_feat.cuda(), words_len.cuda(), weights.cuda(), words_id.cuda(), frames_len=torch.full((1,), 256).cuda())
        
        width = output['width']
        center = output['center']
        selected_props = torch.stack([torch.clamp(center-width/2, min=0), 
                                      torch.clamp(center+width/2, max=1)], dim=-1)
        selected_props = selected_props.cpu().numpy()
        predicted_video_durations[videos] = (selected_props* durations).tolist()[0]

    if exp_const.inf_mode:
        json.dump(img2boxes, open(exp_const.exp_dir.split('/')[-1] + '_' + data_const_subset + '_' + str(model1.const.model_num) + "_raw.json", "w"))

    GT_DATA = json.load(open(osp.join(ROOT_DIR, f'eval_customize/hcstvg_{data_const_subset}_data_gt.json'), 'r'))
    TRACKER_DATA =  json.load(open(osp.join(ROOT_DIR, f"eval_customize/hcstvg_{data_const_subset}_noun_ind_layer_{exp_const.nl}_processed_eval.json"), "r"))    
    TEMPORAL_GRND = predicted_video_durations
    scores = hcstvg_eval_new_anns_v2_temporal(TRACKER_DATA, img2boxes, TEMPORAL_GRND, GT_DATA)
    cls_score = round((success_count/len(dataset)) * 100, 2)
    scores.update({'cls': cls_score})
    return scores



def main(exp_const,data_const,model_const,temp_config=None):
    np.random.seed(exp_const.seed)
    torch.manual_seed(exp_const.seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    
    print('Creating network ...')
    model1 = Constants()
    model1.const = model_const
    model1.object_encoder = ObjectEncoder(model1.const.object_encoder)
    model1.cap_encoder = CapEncoder(model1.const.cap_encoder)
    
    model2 = Constants()
    model2.temp_grounding = temp_reconstruction(temp_config)
    
    o_dim = model1.object_encoder.const.object_feature_dim
    if exp_const.contextualize==True:
        o_dim = model1.object_encoder.const.context_layer.hidden_size
    
    if exp_const.clip:
        model1.cap_encoder.model.config.hidden_size = 512
    print("Spatial contrastive loss....")
    model1.lang_sup_criterion = create_cap_info_nce_criterion(
        o_dim,
        model1.object_encoder.const.object_feature_dim,
        model1.cap_encoder.model.config.hidden_size,
        model1.cap_encoder.model.config.hidden_size//2,
        model1.const.cap_info_nce_layers)
        
    if model1.const.model_num != -1:
        
        loaded_object_encoder = torch.load(model1.const.object_encoder_path)
        print('Loaded model1 number:',loaded_object_encoder['step'])
    
        model1.object_encoder.load_state_dict(
            loaded_object_encoder['state_dict'])
        model1.lang_sup_criterion.load_state_dict(
            torch.load(model1.const.lang_sup_criterion_path)['state_dict'])
                 
        state_dict = torch.load(model_const.temp_ground_path)
        num_updates = state_dict['num_updates']
        parameters = state_dict['model_parameters']
        model2.temp_grounding.load_state_dict(parameters)
        print('load model from {}, num_updates {}.'.format(model_const.temp_ground_path, num_updates))
        
    model1.cap_encoder.cuda()
    model1.object_encoder.cuda()
    # lang criterion
    model1.lang_sup_criterion.cuda()
    model1.cap2_encoder.cuda()
    
    model2.temp_grounding.cuda()
    
    print('Creating dataloader ...')
    dataset = HCSTVGDataset(data_const)

    with torch.no_grad():
        print("weakly supervised evaluation....")
        print("Evaluation...")
        # Evaluate baseline - Whole caption
        results = eval_model_lang_only_temporal(model1, model2, dataset, exp_const, data_const.subset)

    filename = os.path.join(
        exp_const.exp_dir,
        f'results_{data_const.subset}_{model_const.model_num}.json')
    io.dump_json_object(results,filename)
    
