import os
import random
from tqdm import tqdm
import torch
from torch.utils.data import DataLoader
import torch.optim as optim
from torch.utils.tensorboard import SummaryWriter
import numpy as np

import utils.io as io
from utils.constants import save_constants, Constants
from .models.object_encoder import ObjectEncoder
from .models.cap_encoder import CapEncoder
from .models.factored_cap_info_nce_loss_videos import (CapInfoNCE_v2, KLayer, FLayer)
from .dataset_hcstvg import HCSTVGDataset

# temporal localization imports
from .models.temp_reconstruct import temp_reconstruction
from .optimizers.adam_optimizer import AdamOptimizer
from .optimizers.lr_schedulers.inverse_square_root_schedule import InverseSquareRootSchedule

def create_cap_info_nce_criterion(o_dim,u_dim,w_dim,d,layers):
    fo = FLayer(o_dim,d,layers)
    fw = FLayer(w_dim,d,layers)
    ku = KLayer(u_dim,d,layers)
    kw = KLayer(w_dim,d,layers)
    criterion = CapInfoNCE_v2(fo,fw,ku,kw)
    return criterion


def calculate_IoU_batch2(i0, i1):
    union = (np.min(np.stack([i0[0], i1[0]], 0), 0), np.max(np.stack([i0[1], i1[1]], 0), 0))
    inter = (np.max(np.stack([i0[0], i1[0]], 0), 0), np.min(np.stack([i0[1], i1[1]], 0), 0))
    # iou = 1.0 * (inter[1] - inter[0] + 1) / (union[1] - union[0] + 1)
    iou = 1.0 * (inter[1] - inter[0] + 1e-10) / (union[1] - union[0] + 1e-10)
    iou[union[1] - union[0] < -1e-5] = 0
    iou[iou < 0] = 0.0
    return iou

def top_1_metric(pred, label):
    result = {}
    bsz = pred.shape[0]
    iou = calculate_IoU_batch2((pred[:, 0], pred[:, 1]), (label[:, 0], label[:, 1]))
    result['mIoU'] = np.mean(iou)
    for i in range(1, 10, 2):
        result['IoU@0.{}'.format(i)] = 1.0 * np.sum(iou >= i / 10) / bsz
    return result


 
def train_model_st(model1, model2, dataloaders, exp_const, tb_writer):
    params = [
        {'params': model1.object_encoder.parameters()},
        {'params': model1.lang_sup_criterion.parameters()},
    ]
    
    # spatial optimizers
    if exp_const.optimizer == 'SGD':
        opt1 = optim.SGD(
            params,
            lr=exp_const.lr,
            momentum=exp_const.momentum)
    elif exp_const.optimizer == 'Adam':
        opt1 = optim.Adam(
            params,
            lr=exp_const.lr)
    else:
        assert(False), 'optimizer not implemented'
    
    # Temporal optimizers
    model2.temp_grounding.froze_mask_generator()
    parameters = list(filter(lambda p: p.requires_grad, model2.temp_grounding.parameters()))
    args = {'lr': 0.0004, 'weight_decay': 0, 'warmup_updates': 400, 'warmup_init_lr': 1e-07}
    # args['train']["reconstructor"]
    rec_optimizer = AdamOptimizer(args, parameters)
    rec_lr_scheduler = InverseSquareRootSchedule(args, rec_optimizer)

    model2.temp_grounding.froze_reconstructor()
    parameters = list(filter(lambda p: p.requires_grad, model2.temp_grounding.parameters()))
    args = {'lr': 0.0004, 'weight_decay': 0, 'warmup_updates': 400, 'warmup_init_lr': 1e-07}
    mask_optimizer = AdamOptimizer(args, parameters)
    mask_lr_scheduler = InverseSquareRootSchedule(args, mask_optimizer)

    if model1.const.model_num==-1:
        step = 0
    else:
        step = model1.const.model_num

    best_val_loss = 10000
    for epoch in tqdm(range(exp_const.num_epochs), total=exp_const.num_epochs):
        for it,data in tqdm(enumerate(dataloaders['train']), total=len(dataloaders['train'])):
            # Set mode
            model1.lang_sup_criterion.train()
            if exp_const.random_lang is True:
                model1.cap_encoder.train()
            
            # Forward pass
            # spatial data
            object_features = data['tube_feat'].to(torch.float32).cuda() # BxN(objects)xN_FRAMESxFEAT_DIM (2048)
            tube_context_object_features = object_features
                
            # Compute cap-image loss
            with torch.no_grad():
                # token_len - max(25, all_caption_lengths)
                token_ids, tokens, token_lens = model1.cap_encoder.tokenize_batch(
                    data['caption'], max_len=25)
                # BSx Max_token_length -> All caption in the batch
                token_ids = torch.LongTensor(token_ids).cuda()

                # BSXmax_token_lenx768 - token_feats
                token_features, word_word_att = model1.cap_encoder(token_ids)
                
                # Filter out first Noun-adjective only
                if exp_const.nav or exp_const.both:
                    noun_adj_token_ids = data['noun_adj_token_ids'].cuda()
                    nav_token_features, nav_token_mask = model1.cap_encoder.select_embed(
                        token_features,
                        noun_adj_token_ids)
            
            token_mask = torch.zeros(token_ids.size()).cuda()
            if exp_const.nav:
                # calculate the loss for noun-adj-verb
                lang_sup_loss_adj_nav, _, _ = \
                    model1.lang_sup_criterion(
                        tube_context_object_features,
                        object_features,
                        nav_token_features,
                        nav_token_mask,
                        temp_attn=exp_const.temp_attn,
                        num_objects=data['num_objects'])
                if exp_const.both:
                    lang_sup_loss_adj_caption, _, _ = \
                        model1.lang_sup_criterion(
                            tube_context_object_features,
                            object_features,
                            token_features,
                            token_mask,
                            temp_attn=exp_const.temp_attn,
                            num_objects=data['num_objects'])
                    lang_sup_loss_adj = (lang_sup_loss_adj_nav + lang_sup_loss_adj_caption)
                else:
                    lang_sup_loss_adj = lang_sup_loss_adj_nav
            else:
                lang_sup_loss_adj, _, _ = \
                    model1.lang_sup_criterion(
                        tube_context_object_features,
                        object_features,
                        token_features,
                        token_mask,
                        temp_attn=exp_const.temp_attn,
                        num_objects=data['num_objects'])
            lang_sup_loss = lang_sup_loss_adj  
            
            loss = exp_const.lang_sup_loss_wt*lang_sup_loss

            # Backward pass
            opt1.zero_grad()
            loss.backward()
            opt1.step()
            
            # temporal data
            frames_feat = data['frames_feat'].cuda()
            words_feat, words_len, weights, words_id = model1.cap_encoder.tg_tokenize_batch(data['raw_caption'])
            model2.temp_grounding.froze_mask_generator()
            rec_optimizer.zero_grad()
            output = model2.temp_grounding(frames_feat, words_feat.cuda(), words_len.cuda(), weights.cuda(), words_id.cuda(), frames_len=torch.full((frames_feat.shape[0],), 256).cuda())

            loss1, loss_dict = rec_loss(**output)
            
            loss1.backward(retain_graph=True)
            torch.nn.utils.clip_grad_norm_(model2.temp_grounding.parameters(), 10)
            rec_optimizer.step()

            model2.temp_grounding.froze_reconstructor()
            mask_optimizer.zero_grad()
            output = model2.temp_grounding(frames_feat, words_feat.cuda(), words_len.cuda(), weights.cuda(), words_id.cuda(), frames_len=torch.full((frames_feat.shape[0],), 256).cuda())
            # intra
            loss1, ivc_loss_dict = ivc_loss(**output)
            loss_dict.update(ivc_loss_dict)
            
            loss1.backward()
            torch.nn.utils.clip_grad_norm_(model2.temp_grounding.parameters(), 10)
            mask_optimizer.step()

            rec_lr_scheduler.step_update(step)
            mask_lr_scheduler.step_update(step)
            
            if step%exp_const.log_step==0:
                log_items = {
                    'Lang_Sup_Loss/Train': lang_sup_loss.item(),
                    'Loss/Train': loss.item(),
                    'Lr': exp_const.lr,
                    "TG_rec_loss": loss_dict['final_loss'],
                    "TG_ivc_loss": ivc_loss_dict['ivc_loss'],
                }

                log_str = f'Epoch: {epoch} | Iter: {it} | Step: {step} | '
                for name,value in log_items.items():
                    log_str += '{}: {:.5f} | '.format(name,value)
                    tb_writer.add_scalar(name,value,step)

                print(log_str)
            
            if step%(50*exp_const.log_step)==0:
                print(f'Experiment: {exp_const.exp_name}')
                
            if step%exp_const.model_save_step==0:
                save_items = {
                    'object_encoder': model1.object_encoder,
                    'lang_sup_criterion': model1.lang_sup_criterion,
                }
                if exp_const.random_lang is True:
                    save_items['cap_encoder'] = model1.cap_encoder

                for name,nn_model in save_items.items():
                    model_path = os.path.join(
                        exp_const.model_dir,
                        f'{name}_{step}')
                    torch.save({
                        'state_dict': nn_model.state_dict(),
                        'step': step,
                        'val_loss': None,
                        'prev_best_val_loss': best_val_loss},
                        model_path)
                temp_model_path = os.path.join(
                    exp_const.model_dir,
                    f'temp_model_{step}.pt')
                state_dict = {
                        'num_updates': step,
                        'model_parameters': model2.temp_grounding.state_dict(),
                }
                torch.save(state_dict, temp_model_path)

            if step%exp_const.val_step==0 and exp_const.dataset=='hcstvgv1':
                with torch.no_grad():
                    eval_results = eval_model_contrastive(
                        model1,
                        dataloaders['val'],
                        exp_const,
                        step)

                print('Val reults:',eval_results)
                log_items = {
                    'Lang_Sup_Loss/Val': eval_results['lang_sup_loss'],
                    'Loss/Val': eval_results['total_loss']
                }
                
                for name,value in log_items.items():
                    tb_writer.add_scalar(name,value,step)

                val_loss = eval_results['total_loss']
                if val_loss < best_val_loss:                    
                    print(f'Saving best model1 at {step} ...')
                    save_items = {
                        'best_object_encoder': model1.object_encoder,
                        'best_lang_sup_criterion': model1.lang_sup_criterion,
                    }
                    if exp_const.random_lang is True:
                        save_items['best_cap_encoder'] = model1.cap_encoder

                    for name,nn_model in save_items.items():
                        model_path = os.path.join(
                            exp_const.model_dir,
                            name)
                        torch.save({
                            'state_dict':nn_model.state_dict(),
                            'step': step,
                            'val_loss': val_loss,
                            'prev_best_val_loss': best_val_loss},
                            model_path)

                    best_val_loss = val_loss
            step += 1
 

def eval_model_contrastive(model1,dataloader,exp_const,step):
    # Set mode
    model1.object_encoder.eval()
    model1.lang_sup_criterion.eval()
    if exp_const.random_lang is True:
        model1.cap_encoder.eval()

    avg_lang_sup_loss = 0
    num_samples = 0
    for it,data in enumerate(tqdm(dataloader, total=len(dataloader))):
        if (exp_const.num_val_samples is not None) and \
            (num_samples >= exp_const.num_val_samples):
                break
        
        # Forward pass
        object_features = data['tube_feat'].to(torch.float32).cuda()
        object_mask = data['object_mask'].cuda()
        pad_mask = data['tube_pad_mask'].cuda()

        if exp_const.contextualize==True:
            tube_context_object_features, _ = model1.object_encoder(
                object_features,
                object_mask,
                pad_mask)
        else:
            tube_context_object_features = object_features 

        token_ids, tokens, token_lens = model1.cap_encoder.tokenize_batch(
            data['caption'], max_len=25)
        
        # BSx Max_token_length -> All caption in the batch
        token_ids = torch.LongTensor(token_ids).cuda()
        
        # BSXmax_token_lenx768 - token_feats
        token_features, word_word_att = model1.cap_encoder(token_ids)
        token_mask = torch.zeros(token_ids.size()).cuda()
        
        if exp_const.nav:
            noun_adj_token_ids = data['noun_adj_token_ids'].cuda()
            nav_token_features, nav_token_mask = model1.cap_encoder.select_embed(
                token_features,
                noun_adj_token_ids)    
        
        
            lang_sup_loss_adj_nav, _, _ = model1.lang_sup_criterion(
                tube_context_object_features,
                object_features,
                nav_token_features,
                nav_token_mask,
                temp_attn=exp_const.temp_attn)
            if exp_const.both:
                lang_sup_loss_adj_caption, _, _ = model1.lang_sup_criterion(
                    tube_context_object_features,
                    object_features,
                    token_features,
                    token_mask,
                    temp_attn=exp_const.temp_attn)
                lang_sup_loss_adj = lang_sup_loss_adj_nav + lang_sup_loss_adj_caption
            else:
                lang_sup_loss_adj = lang_sup_loss_adj_nav
        else:
            lang_sup_loss_adj, _, _ = model1.lang_sup_criterion(
                tube_context_object_features,
                object_features,
                token_features,
                token_mask,
                temp_attn=exp_const.temp_attn)
        lang_sup_loss = lang_sup_loss_adj
        
        # Aggregate loss or accuracy
        batch_size = object_features.size(0)
        num_samples += batch_size
        avg_lang_sup_loss += (lang_sup_loss.item()*batch_size)

    avg_lang_sup_loss = avg_lang_sup_loss / num_samples
    total_loss =  exp_const.lang_sup_loss_wt*avg_lang_sup_loss

    eval_results = {
        'lang_sup_loss': avg_lang_sup_loss,
        'total_loss': total_loss,
    }
    return eval_results


def main(exp_const,data_const,model_const, temp_config=None):
    random.seed(exp_const.seed)
    np.random.seed(exp_const.seed)
    torch.manual_seed(exp_const.seed)
    torch.cuda.manual_seed_all(exp_const.seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False

    io.mkdir_if_not_exists(exp_const.exp_dir,recursive=True)
    io.mkdir_if_not_exists(exp_const.log_dir)
    io.mkdir_if_not_exists(exp_const.model_dir)
    io.mkdir_if_not_exists(exp_const.vis_dir)
    
    tb_writer = SummaryWriter(log_dir=exp_const.log_dir)
    
    model_num = model_const.model_num
    save_constants({
        f'exp_{model_num}': exp_const,
        f'data_train_{model_num}': data_const['train'],
        f'data_val_{model_num}': data_const['val'],
        f'model_{model_num}': model_const},
        exp_const.exp_dir)
    
    print('Creating network ...')
    # spatial constants
    model1 = Constants()
    # temporal constants
    model2 = Constants()
    
    model1.const = model_const
    model1.object_encoder = ObjectEncoder(model1.const.object_encoder)
    
    # Language encoder - Bert
    model1.cap_encoder = CapEncoder(model1.const.cap_encoder)
    model2.temp_grounding = temp_reconstruction(temp_config)
    
    o_dim = model1.object_encoder.const.object_feature_dim
    
    print("Spatial contrastive loss....")
    model1.lang_sup_criterion = create_cap_info_nce_criterion(
        o_dim,
        model1.object_encoder.const.object_feature_dim,
        model1.cap_encoder.model.config.hidden_size,
        model1.cap_encoder.model.config.hidden_size//2,
        model1.const.cap_info_nce_layers)
    
    # spatial training
    if model1.const.model_num != -1:
        print("Loading previous weights for Spatial training...")
        model1.object_encoder.load_state_dict(
            torch.load(model1.const.object_encoder_path)['state_dict'])
        model1.lang_sup_criterion.load_state_dict(
            torch.load(model1.const.lang_sup_criterion_path)['state_dict'])
        print("Spatial model loaded...")
        
        print("Loading previous weights for Temporal training...")
        state_dict = torch.load(model_const.temp_ground_path)
        num_updates = state_dict['num_updates']
        parameters = state_dict['model_parameters']
        model2.temp_grounding.load_state_dict(parameters)
        print('load model from {}, num_updates {}.'.format(model_const.temp_ground_path, num_updates))
        print("Temporal model loaded...")

    model1.object_encoder.cuda()
    model1.cap_encoder.cuda()
    model1.lang_sup_criterion.cuda()
    
    model1.cap2_encoder.cuda()
    
    model1.lang_sup_criterion.to_file(
        os.path.join(exp_const.exp_dir,'lang_supervised_criterion.txt'))
        
    model1.object_encoder.to_file(
        os.path.join(exp_const.exp_dir,'object_encoder.txt'))
    
    # temporal grounding
    model2.temp_grounding.cuda()
    
    print('Creating dataloader ...')
    dataloaders = {}
    if exp_const.dataset=='hcstvg':
        Dataset = HCSTVGDataset
    elif exp_const.dataset=='hcstvgv2':
        Dataset = HCSTVGv2Dataset
    elif exp_const.dataset=='vidstg':
        Dataset = VidSTGDataset
    else:
        msg = f'{exp_const.dataset} not implemented'
        raise NotImplementedError(msg)

    def worker_init_fn(worker_id):
        def set_seed(seed):
            import random
            import numpy as np
            import torch

            random.seed(seed)
            np.random.seed(seed)
            torch.manual_seed(seed)
            torch.cuda.manual_seed(seed)
            torch.cuda.manual_seed_all(seed)

        set_seed(exp_const.seed + worker_id)
            
    for mode, const in data_const.items():
        dataset = Dataset(const)
        
        if mode=='train':
            shuffle=True
            batch_size=exp_const.train_batch_size
            worker_init_fn = worker_init_fn
            dataloaders[mode] = DataLoader(
                dataset,
                batch_size=batch_size,
                shuffle=shuffle,
                num_workers=exp_const.num_workers,
                drop_last=True,
                worker_init_fn=worker_init_fn)
        else:
            shuffle=False
            batch_size=exp_const.val_batch_size

            dataloaders[mode] = DataLoader(
                dataset,
                batch_size=batch_size,
                shuffle=shuffle,
                num_workers=exp_const.num_workers,
                drop_last=True)

    print("Weakly supervised training...")
    train_model_st(model1, model2, dataloaders, exp_const, tb_writer)
    